#include "mpi.h"
#include <string>
#include <iostream>
#include <fstream>
#include <sstream>
#include <vector>
#include <getopt.h>
#include <assert.h>
#include "Uncompress.h"
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <fcntl.h>
#include <algorithm>
#include <queue>
# include <omp.h>
#include "partition.h"

#define PROGRAM "dida_mpi"
#define READ 0
#define WRITE 1


static const char VERSION_MESSAGE[] =
PROGRAM " Version 0-1.2 \n"
"Written by Hamid Mohamadi.\n"
"Copyright 2013 Canada's Michael Smith Genome Science Centre\n";

static const char USAGE_MESSAGE[] =
"Usage: " PROGRAM " [OPTION]... QUERY... TARGET\n"
"Dispatch the sequences of the files QUERY based on the Bloom filter of the file TARGET.\n"
"The index files TARGET.fai and TARGET.fm will be used if present.\n"
"\n"
" Options:\n"
"\n"
"  -p, --partition=N       divide reference to N partitions\n"
"  -j, --threads=N         use N parallel threads [partitions]\n"
"  -b, --bfl=N             use N bp for Bloom filter windows[1]\n"
"  -h, --hash=N            use N hash functions for Bloom filter [6]\n"
"      --help              display this help and exit\n"
"      --version           output version information and exit\n"
"\n"
"Report bugs to hmohamadi@bcgsc.ca.\n";

namespace opt {
	/** Number of bits per item. */
	unsigned ibits = 8;
    
	/** The number of parallel threads. */
	static unsigned threads = 0;
    
	/** The number of partitions. */
	static int pnum = 1;
    
	/** The number of hash functions. */
	int nhash = 5;
    
	/** The size of a k-mer. */
	int bmer = 20;
}

static const char shortopts[] = "p:b:j:h:";

enum { OPT_HELP = 1, OPT_VERSION };

static const struct option longopts[] = {
	{ "threads",	required_argument, NULL, 'j' },
	{ "partition",	required_argument, NULL, 'p' },
	{ "bfl",	required_argument, NULL, 'b' },
	{ "hash",	required_argument, NULL, 'h' },
	{ "help",	no_argument, NULL, OPT_HELP },
	{ "version",	no_argument, NULL, OPT_VERSION },
	{ NULL, 0, NULL, 0 }
};

/*void getFname(const char *filename, std::string &bName, std::string &eName){
	std::string fName(filename);
	size_t pos = fName.rfind(".");
	if(pos == std::string::npos){
		bName = fName;
		eName = "";
		return;
	}
	if(pos == 0){
		bName = "";
		eName = fName.substr(pos+1, std::string::npos);
		return;
	}
	bName = fName.substr(0, pos);
	eName = fName.substr(pos+1, std::string::npos);
}*/

size_t getInfo(const char *aName, unsigned k){
	std::string line;
	std::ifstream faFile(aName);
	size_t totItm=0;
	while(getline(faFile, line)){
		getline(faFile, line);
		size_t uLen = line.length();
		totItm+=uLen-k+1;
	}
	std::cerr << "|totLen|=" << totItm << std::endl;
	faFile.close();
	return totItm;
}

// MurmurHash2, 64-bit versions, by Austin Appleby
// https://sites.google.com/site/murmurhash/MurmurHash2_64.cpp?attredirects=0
uint64_t MurmurHash64A ( const void * key, int len, unsigned int seed ){
	const uint64_t m = 0xc6a4a7935bd1e995;
	const int r = 47;
    
	uint64_t h = seed ^ (len * m);
    
	const uint64_t * data = (const uint64_t *)key;
	const uint64_t * end = data + (len/8);
    
	while(data != end)
	{
		uint64_t k = *data++;
        
		k *= m;
		k ^= k >> r;
		k *= m;
		
		h ^= k;
		h *= m;
	}
    
	const unsigned char * data2 = (const unsigned char*)data;
    
	switch(len & 7)
	{
        case 7: h ^= uint64_t(data2[6]) << 48;
        case 6: h ^= uint64_t(data2[5]) << 40;
        case 5: h ^= uint64_t(data2[4]) << 32;
        case 4: h ^= uint64_t(data2[3]) << 24;
        case 3: h ^= uint64_t(data2[2]) << 16;
        case 2: h ^= uint64_t(data2[1]) << 8;
        case 1: h ^= uint64_t(data2[0]);
	        h *= m;
	};
    
	h ^= h >> r;
	h *= m;
	h ^= h >> r;
    
	return h;
}

void filInsert(std::vector< std::vector<bool> > &myFilters, const unsigned pn, const std::string &bMer) {
	for (int i=0; i < opt::nhash; ++i)
		myFilters[pn][MurmurHash64A(bMer.c_str(), opt::bmer, i) % myFilters[pn].size()]=true;
}

bool filContain(const std::vector< std::vector<bool> > &myFilters, const unsigned pn, const std::string &bMer) {
	for (int i=0; i < opt::nhash; ++i)
		if (!myFilters[pn][MurmurHash64A(bMer.c_str(), opt::bmer, i) % myFilters[pn].size()])
			return false;
	return true;
}

static inline char rc(char c) {
    switch (c) {
        case 'A':
            return 'T';
            break;
        case 'C':
            return 'G';
            break;
        case 'G':
            return 'C';
            break;
        case 'T':
            return 'A';
            break;
        default:
            break;
    }
    return c;
}

struct samHed
{
    std::string SQ1;
	char sn1;
    char sn2;
    char sn3;
	unsigned sqId;
    std::string SQ3;
    int hedPr;
};

struct samRec
{
    unsigned SamOrd;
    std::string SamQn;
    int SamFg;
    std::string SamRf;
    int SamPs;
    int SamMq;
    std::string SamCr;
    std::string SamRn;
    int SamPn;
    int SamTl;
    std::string SamSq;
    std::string SamPh;
    int SamPr;
};

bool operator>(const samHed& lhs, const samHed& rhs){
    return lhs.sqId > rhs.sqId;
}

samHed hedLoad(std::string& line, int partNum) {
    std::istringstream iss(line);
    samHed c;
    iss>>c.SQ1>>c.sn1>>c.sn2>>c.sn3>>c.sqId>>c.SQ3;
    c.hedPr=partNum;
    return c;
}

std::ostream& operator<<(std::ostream& os, const samHed& c){
    os<<c.SQ1<<"\t"<<c.sn1<<c.sn2<<c.sn3<<c.sqId<<"\t"<<c.SQ3;
    return os;
}

bool operator>(const samRec& lhs, const samRec& rhs){
    return lhs.SamOrd > rhs.SamOrd;
}

samRec recLoad(std::string& line, int partNum){
    std::istringstream iss(line);
    samRec c;
    char hSep;
    iss>>c.SamOrd>>hSep>>c.SamQn>>c.SamFg>>c.SamRf>>c.SamPs>>c.SamMq>>c.SamCr;
    c.SamRn="*";c.SamPn=0;c.SamTl=0;c.SamSq="*";c.SamPh="*";
    c.SamPr=partNum;
    return c;
}

std::ostream& operator<<(std::ostream& os, const samRec& c){
    os<<c.SamQn<<"\t"<<c.SamFg<<"\t"<<c.SamRf<<"\t"<<c.SamPs<<"\t"<<c.SamMq<<"\t"<<c.SamCr<<"\t*\t0\t0\t*\t*";
    return os;
}

int main(int argc, char** argv) {
    
    bool die = false;
    std::string blPath;
    
	for (int c; (c = getopt_long(argc, argv, shortopts, longopts, NULL)) != -1;) {
		std::istringstream arg(optarg != NULL ? optarg : "");
		switch (c) {
			case '?':
				die = true; break;
			case 'j':
				arg >> opt::threads; break;
			case 'b':
				arg >> opt::bmer; break;
			case 'p':
				arg >> opt::pnum; break;
			case 'h':
				arg >> opt::nhash; break;
                
			case OPT_HELP:
				std::cerr << USAGE_MESSAGE;
				exit(EXIT_SUCCESS);
			case OPT_VERSION:
				std::cerr << VERSION_MESSAGE;
				exit(EXIT_SUCCESS);
		}
		if (optarg != NULL && !arg.eof()) {
			std::cerr << PROGRAM ": invalid option: `-"
			<< (char)c << optarg << "'\n";
			exit(EXIT_FAILURE);
		}
	}
    
	if (opt::bmer == 0) {
		std::cerr << PROGRAM ": missing mandatory option `-b'\n";
		die = true;
	}
    
	if (argc - optind != 2) {
		std::cerr << PROGRAM ": missing arguments\n";
		die = true;
	}
    
	if (die) {
		std::cerr << "Try `" << PROGRAM
		<< " --help' for more information.\n";
		exit(EXIT_FAILURE);
	}
    
	const char *libName(argv[argc-2]);
	const char *refName(argv[argc-1]);
    
    int procSize, procRank, prcrNlen, provided;
	char processor_name[MPI_MAX_PROCESSOR_NAME];
    
	MPI_Status status;
    MPI_Init_thread(&argc, &argv, MPI_THREAD_MULTIPLE, &provided);
	MPI_Comm_size(MPI_COMM_WORLD, &procSize);
	MPI_Comm_rank(MPI_COMM_WORLD, &procRank);
	MPI_Get_processor_name(processor_name,&prcrNlen);

	printf("process %d out of %d on %s with thread level %d\n", procRank, procSize, processor_name, provided);

    std::cerr << refName << "\n";
    
    if (procRank==0)
            getPartition(refName, opt::pnum);

    MPI_Barrier(MPI_COMM_WORLD);

    if (procRank > 0 && procRank < procSize-1) {
            std::stringstream indstm;
            std::cout << "running command: abyss-index mref-" << procRank << ".fa\n";
            indstm << "abyss-index mref-" << procRank << ".fa";
            std::system(indstm.str().c_str());
            //indstm << "mref-" << procRank << ".fa";
    		//execlp("abyss-index", "abyss-index", indstm.str().c_str(), (char *)0);
        }
    MPI_Barrier(MPI_COMM_WORLD);

    int readLen=0, recLen=0;
    // BEGIN MPI MASTER PROCESS LOAD FILTER
   	if(procRank==0){
        std::cerr << "Number of hash functions=" << opt::nhash << "\n";
        
        unsigned tNum = omp_get_max_threads()>opt::pnum?opt::pnum:omp_get_max_threads();
        if (opt::threads < tNum && opt::threads > 0)
            tNum = opt::threads;
        std::cerr << "Number of threads=" << tNum << std::endl;
        omp_set_num_threads(tNum);
        
        int pIndex,chunk=1;
        //begin create filters
        std::vector< std::vector<bool> > myFilters(opt::pnum);
        
        std::cerr << "Loading filters ...\n";
#pragma omp parallel for shared(myFilters) private(pIndex) schedule(static,chunk)
        for (pIndex=0; pIndex<opt::pnum; ++pIndex){
            std::stringstream sstm;
            sstm << "mref-" << pIndex+1 << ".fa";
            size_t filterSize = opt::ibits*getInfo((sstm.str()).c_str(), opt::bmer);
            myFilters[pIndex].resize(filterSize);
            
            std::ifstream uFile(sstm.str().c_str());
            std::string line;
            while (getline(uFile, line)){
                getline(uFile, line);
                std::transform(line.begin(), line.end(), line.begin(), ::toupper);
                long uL= line.length();
                for(long j = 0; j < uL -opt::bmer+1; ++j) {
                    std::string bMer = line.substr(j,opt::bmer);
                    //Begin
                    int p=0,hLen=(opt::bmer-1)/2;
                    while(bMer[p] == rc(bMer[opt::bmer-1-p])){
                        ++p;
                        if(p>hLen)break;
                    }
                    if (bMer[p] > rc(bMer[opt::bmer-1-p])) {
                        for (int lIndex = p, rIndex = opt::bmer-1-p; lIndex<rIndex; ++lIndex,--rIndex) {
                            char tmp = rc(bMer[rIndex]);
                            bMer[rIndex] = rc(bMer[lIndex]);
                            bMer[lIndex] = tmp;
                        }
                    }
                    //End
                    filInsert(myFilters,pIndex,bMer);
                }
            }
            uFile.close();
        }
        
        std::cerr<< "Loading BF done!" << std::endl;
        
        std::string readHead, readSeq, readDir, readQual, rName;
        unsigned readId=0;
        
        std::ifstream libFile(libName);
        while(getline(libFile, rName)) {
            std::ifstream readFile(rName.c_str());
            while(getline(readFile, readHead)) {
                getline(readFile, readSeq);
                std::transform(readSeq.begin(), readSeq.end(), readSeq.begin(), ::toupper);
                getline(readFile, readDir);
                getline(readFile, readQual);
                readLen = readSeq.length();
                std::stringstream hstm;
                //hstm << ">" << (readId);
                //std::string readRec = hstm.str() + readSeq;
                readHead[0]=':'; hstm << ">" << readId << readHead;
                std::string readRec = hstm.str() + readSeq;
                //readHead[0]='>';  std::string readRec = readHead + readSeq;
                recLen = readRec.length();
#pragma omp parallel for shared(myFilters) private(pIndex) schedule(static,chunk)
                for(pIndex=1; pIndex<=opt::pnum; ++pIndex)
                    for(int j=0; j<readLen-opt::bmer+1; j+=opt::bmer){
                        std::string bMer = readSeq.substr(j,opt::bmer);
                        //Begin
                        int p=0,hLen=(opt::bmer-1)/2;
                        while(bMer[p] == rc(bMer[opt::bmer-1-p])){
                            ++p;
                            if(p>hLen)break;
                        }
                        if (bMer[p] > rc(bMer[opt::bmer-1-p])) {
                            for (int lIndex = p, rIndex = opt::bmer-1-p; lIndex<rIndex; ++lIndex,--rIndex) {
                                char tmp = rc(bMer[rIndex]);
                                bMer[rIndex] = rc(bMer[lIndex]);
                                bMer[lIndex] = tmp;
                            }
                        }
                        //END
                        if (filContain(myFilters, pIndex-1, bMer)) {
                            // MPI SEND READ
                            MPI_Send(&readLen, 1, MPI_INT, pIndex, 0, MPI_COMM_WORLD);
                            MPI_Send(&recLen, 1, MPI_INT, pIndex, 0, MPI_COMM_WORLD);
                            MPI_Send(&readRec[0], recLen, MPI_CHAR, pIndex, 0, MPI_COMM_WORLD);
                            break;
                        }
                        
                    }
                ++readId;
                
            }
            readFile.close();
        }
        libFile.close();
        int minImpi = -1;
        for(int pIndex = 1; pIndex<=opt::pnum; ++pIndex)
            MPI_Send(&minImpi, 1, MPI_INT, pIndex, 0, MPI_COMM_WORLD);
    }
    // END MPI MASTER PROCESS LOAD FILTER
//    else {
     else if (procRank < procSize-1 && procRank > 0) {
        int readLen=0, recLen=0;
        int pid, fd[2];
        int fd2[2];
        if (pipe(fd) == -1 || pipe(fd2) == -1) {
            perror("pipe failed");
            exit(1);
        }
        
        if ((pid = fork()) < 0) {
            perror("fork failed");
            exit(2);
        }
        
        if(pid != 0) {
            close(fd[READ]);
            dup2(fd[WRITE],1);
            close(fd[WRITE]) ;
            close(fd2[WRITE]);
            dup2(fd2[READ],0);
            close(fd2[READ]);
            #pragma omp parallel sections
            {
                #pragma omp section
                {
                    for(;;) {
                        MPI_Recv(&readLen, 1, MPI_INT, 0, 0, MPI_COMM_WORLD, &status);
                        if(readLen==-1) break;
                        MPI_Recv(&recLen, 1, MPI_INT, 0, 0, MPI_COMM_WORLD, &status);
                        char * readbuf = new char [recLen + 1];
                        MPI_Recv(readbuf,recLen,MPI_CHAR,0,0,MPI_COMM_WORLD,&status);
                        
                        std::string myHead(readbuf,readbuf+recLen-readLen);
                        std::string myRead(readbuf+recLen-readLen,readbuf+recLen);
                        delete [] readbuf;
                        fprintf(stdout, "%s\n%s\n", myHead.c_str(), myRead.c_str());
                        
                    }
                    close(STDOUT_FILENO);
                }
                #pragma omp section
                {
                    char *line = NULL;
                    size_t lineByte = 0;
                    ssize_t lineLen;
                    while ((lineLen = getline(&line, &lineByte, stdin)) != -1) {
                        MPI_Send(&lineLen, 1, MPI_INT, procSize-1, 0, MPI_COMM_WORLD);
                        MPI_Send(line, lineLen, MPI_CHAR, procSize-1, 0, MPI_COMM_WORLD);
                    }
                    free(line);
                    lineLen=-1;
                    MPI_Send(&lineLen, 1, MPI_INT, procSize-1, 0, MPI_COMM_WORLD);
                }
            }            
        }
        else {
            close(fd[WRITE]);
            dup2(fd[READ], 0);
            close(fd[READ]);
            close(fd2[READ]);
            dup2(fd2[WRITE],1);
            close(fd2[WRITE]);
            std::stringstream amapstm, lstm;
            amapstm << "mref-" << procRank << ".fa";
            lstm << "-l" << opt::bmer;
            execlp("abyss-map", "abyss-map", "--order", "-j10", "-l50", "-", amapstm.str().c_str(), (char *)0);

        }
    }
    else if(procRank==procSize-1) {
        int isize=0;
        char * readbuf;
        std::priority_queue< samHed, std::vector<samHed>, std::greater<samHed> > hedBuffer;
        std::priority_queue< samRec, std::vector<samRec>, std::greater<samRec> > recBuffer;
        for (int pIndex = 1; pIndex <= opt::pnum; ++pIndex) {
            //Skipping first header lines from each sam File
            MPI_Recv(&isize, 1, MPI_INT, pIndex, 0, MPI_COMM_WORLD, &status);
            readbuf = new char [isize];
            MPI_Recv(readbuf,isize,MPI_CHAR,pIndex,0,MPI_COMM_WORLD,&status);
            delete [] readbuf;
        }
        for (int pIndex = 1; pIndex <= opt::pnum; ++pIndex) {
            //Skipping second header lines from each sam File
            MPI_Recv(&isize, 1, MPI_INT, pIndex, 0, MPI_COMM_WORLD, &status);
            readbuf = new char [isize];
            MPI_Recv(readbuf,isize,MPI_CHAR,pIndex,0,MPI_COMM_WORLD,&status);
            delete [] readbuf;
        }
        for (int pIndex = 1; pIndex <= opt::pnum; ++pIndex) {
            MPI_Recv(&isize, 1, MPI_INT, pIndex, 0, MPI_COMM_WORLD, &status);
            readbuf = new char [isize];
            MPI_Recv(readbuf,isize,MPI_CHAR,pIndex,0,MPI_COMM_WORLD,&status);
            std::string sambuf(readbuf, readbuf+isize-1);
            delete [] readbuf;
            hedBuffer.push(hedLoad(sambuf,pIndex));
        }
        std::ofstream comFile("aln.sam");
        while (!hedBuffer.empty()) {
            samHed cHed=hedBuffer.top();
            comFile<<cHed<<"\n";
            hedBuffer.pop();
            MPI_Recv(&isize, 1, MPI_INT, cHed.hedPr, 0, MPI_COMM_WORLD, &status);
            readbuf = new char [isize];
            MPI_Recv(readbuf,isize,MPI_CHAR,cHed.hedPr,0,MPI_COMM_WORLD,&status);
            std::string sambuf(readbuf, readbuf+isize-1);
            delete [] readbuf;
            if(sambuf[0]=='@')
                hedBuffer.push(hedLoad(sambuf,cHed.hedPr));
            else
                recBuffer.push(recLoad(sambuf,cHed.hedPr));
        }
        
        for (int dIndex = 1; dIndex <= 2; ++dIndex)
        for (int pIndex = 1; pIndex <= opt::pnum; ++pIndex) { //guard for -1 from MPI_Recv
            MPI_Recv(&isize, 1, MPI_INT, pIndex, 0, MPI_COMM_WORLD, &status);
            readbuf = new char [isize];
            MPI_Recv(readbuf,isize,MPI_CHAR,pIndex,0,MPI_COMM_WORLD,&status);
            std::string sambuf(readbuf, readbuf+isize-1);
            delete [] readbuf;
            recBuffer.push(recLoad(sambuf,pIndex));
        }
        bool prtDone[opt::pnum];
        for (int pIndex = 1; pIndex <= opt::pnum; ++pIndex)
            prtDone[pIndex-1] = false;
            
        while (!recBuffer.empty()) {
            samRec cRec=recBuffer.top();
            if(cRec.SamFg!= 4)comFile<<cRec<<"\n";
            //comFile<<cRec<<"\n";
            recBuffer.pop();
            
            if(!prtDone[cRec.SamPr-1])
                MPI_Recv(&isize, 1, MPI_INT, cRec.SamPr, 0, MPI_COMM_WORLD, &status);
            else
                isize = -1;
            if(isize!=-1) {
                readbuf = new char [isize];
                MPI_Recv(readbuf,isize,MPI_CHAR,cRec.SamPr,0,MPI_COMM_WORLD,&status);
                std::string sambuf(readbuf, readbuf+isize-1);
                delete [] readbuf;
                recBuffer.push(recLoad(sambuf,cRec.SamPr));
            }
            else
                prtDone[cRec.SamPr-1] = true;
        }
        comFile.close();
    }

	MPI_Finalize();
	return 0;
}
